package chatserver;

import java.io.*;
import java.net.*;
import java.util.*;

/**
This program implements a multithreaded server that listens to port 8189 and echoes back 
all client input.
Copied code from Cay Horstmann
@version 1.20 2004-08-03
 */
public class Server implements Runnable {

    private ArrayList<Client> clients;
    private boolean running = true;

    private static final int PORT_NUMBER = 8189;
    /**
     * Builds a server
     */
    public Server() {
        clients = new ArrayList<Client>();
    }

    public void run() {
        ChatGui.MessageOut("Server created: Up and running");
        try {
            ServerSocket s = new ServerSocket(PORT_NUMBER);
            while (running) {
                Socket incoming = s.accept();  // waits until it gets a socket and then assigns it to incoming
                Client c = new Client(incoming);
                for (Client i : clients) { // tells the new member who all is here
                    this.message(c.getSocket(), i.getName() + " joined the chat.");
                }
                clients.add(c);
                ChatGui.updateList();
                broadcast(c.getName() + " joined the chat.");
                Thread t = new Thread(c);
                t.start();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Broadcasts the text to all the people that are attached to the 
     * @param s the String to broadcast to call connected clients
     * @throws java.io.IOException
     */
    public void broadcast(String s) throws IOException {
        for (Client c : clients) {
            message(c.getSocket(), s);
        }
        ChatGui.MessageOut(s);
    }

    /**
     * Send a private message to a specific name 
     * @param client the name of the client
     * @param s the string to send
     * @throws java.io.IOException
     */
    public void message(String client, String s) throws IOException {
        boolean sent = false;
        for (Client c : clients) {
            if (c.getName().equals(client)) {
                s = "Server to " + client + ": " + s;
                message(c.getSocket(), s);
                ChatGui.MessageOut(s);
                sent = true;
            }
        }
        if (!sent) {
            ChatGui.MessageOut("The Client: " + client + " does not exist.");
        }
    }

    /**
     * A faster Message that knows who to send it to
     * @param soc the socket to send the message on
     * @param s the string to send
     * @throws java.io.IOException
     */
    private void message(Socket soc, String s) throws IOException {
        if (!soc.isClosed()) {
            OutputStream outStream = soc.getOutputStream();
            PrintWriter out = new PrintWriter(outStream, true /* autoFlush */);
            out.println(s);
        }
    }

    /**
     * Handles an arraylist of names to disconnect
     * @param name an arraylist of Strings which disignate the names to disconnect
     * @throws java.io.IOException
     */
    public void disconnect(String name) throws IOException {
        for (Client c : clients) {
            if (c.getName().equals(name)) {
                broadcast(name + " quit with message: Kicked.");
                c.disconnect();
            }
        }
    }

    /**
     * Returns an array of Strings which are the names of clients connected
     * @return the client names
     */
    public String[] clients() {
        String[] list = new String[clients.size()];
        int i = 0;
        for (Client c : clients) {
            list[i] = c.getName();
            i++;
        }
        return list;
    }

    /**
     * Checks all available names to see if "name" is available
     * @param name the name to check
     * @return a valid name that will be used
     */
    public String requestName(String name) {
        // check if it's available
        if (clients.isEmpty()) { // if the name list is empty, DON'T INTERATE THROUGH IT
            return name;
        }
        boolean available = true;
        for (Client c : clients) {
            if (c.getName().equals(name)) {
                available = false;
            }
        }
        if (available) { // name is ok
            return name;
        } else { // find a new name
            String numbers[] = name.split("[a-zA-Z]");
            if (numbers.length < 1) {
                return requestName(name + "1");
            }
            String last = numbers[numbers.length - 1];
            if (last.matches("[0-9]")) { // we have just numbers at the end
                Integer lastnumber = new Integer(last);
                int newnumber = lastnumber + 1;
                return requestName(name.substring(0, name.length() - 1) + newnumber);
            } else { // um. no numbers at the end, let's attach one.
                return requestName(name + "1");
            }
        }
    }

    /**
     * Toggles the state of the server
     * Currently not used. but could be...
     */
    public void toggleServer() {
        if (running) { running = false; }
        else { running = true; }
    }

    /**
     * A class which handles each connection
     */
    class Client implements Runnable {
        private Socket connection; // the individual socket for each client
        private String name; // the assosiated name
        private static final String NAME_CHECK = "[a-zA-Z0-9]"; // the legal characters in the name

        /**
         * The client construction
         * Defaults to Guest on arival, the individual can change the name
         * @param c the scoket to build the client on
         */
        public Client(Socket c) {
            connection = c;
            name = requestName("guest");
        }

        public void run() {
            try {
                try {
                    Scanner in = new Scanner(connection.getInputStream());
                    PrintWriter out = new PrintWriter(connection.getOutputStream(), true /* autoFlush */);
                    out.println("Use '/nick {name}' to change your name\nand use '/quit {message}' to leave.");

                    while (!(connection.isClosed()) && in.hasNextLine()) {
                        String line = in.nextLine();
                        if (line.trim().matches("/(.*)")) { // run a command
                            command(line.trim());
                        } else { // just say it
                            broadcast(header() + line);
                        }
                    }
                } finally {
                    broadcast(name + " quit with message: Disconnected.");
                    this.disconnect();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        /**
         * Parses the line for commands
         * @param line the line to check
         * @throws java.io.IOException
         */
        private void command(String line) throws IOException {
            String items[] = line.split(" ");// split according to the (space) delimeter
            if (items[0].equals("/quit")) { // Broadcast the quit message and disconnect the client
                String aMessage = getName() + " quit with message: ";
                for (int i = 1; i < items.length; i++) {
                    aMessage = aMessage + " " + items[i];
                }
                broadcast(aMessage);
                this.disconnect();
            }
            if (items[0].equals("/nick")) { // changing their name
                if (isValid(items[1])) {
                    setName(items[1], name); // assume they have no trailing characters
                } else { // bad name
                    message(name, "Name is invalid. Only Alphanumerics allowed.");
                }
            } else { // bad command
                message(name, "Command not recognized.");
            }
        }

        /**
         * checks to see if the name is a valid name
         * @param name the name that is being tested
         * @return if the name is only alpha numeric
         */
        public boolean isValid(String name) {
            // split the string with all alphanumeric
            String remaining[] = name.split(NAME_CHECK);
            // if any of the chars remaining are non-null, then the string invalid
            for (int i = 0; i < remaining.length; i++) {
                if (remaining[i] != null) {
                    return false;
                }
            }
            return true;
        }

        /**
         * Returns the connection
         * @return the connection
         */
        public Socket getSocket() {
            return connection;
        }

        /**
         * Returns the name of the client
         * @return the name of the client
         */
        public String getName() {
            return name;
        }

        /**
         * Disconnects the client
         * @throws java.io.IOException
         */
        public void disconnect() throws IOException {
            clients.remove(this);
            ChatGui.updateList();
            connection.close();
        }

        /**
         * Attempts to set the name
         * If the name is taken it simply changes it to a similar valid name
         * @param newName The name that is being requested
         * @param oldName the old name
         * @throws java.io.IOException
         */
        public void setName(String newName, String oldName) throws IOException {
            name = requestName(newName);
            broadcast(oldName + " became " + name);
            ChatGui.updateList();
        }

        /**
         * Shows a clean header before each thing a client says
         * @return the header for the client when talking
         */
        private String header() {
            return name + " : ";  // there must be spaces around the ':' because of the way the text is parsed
        }                               // If it was missing, then people could say "joined" or "quit" 
    }                                  // and a member with their name and a semicolon would join and leave the list of "peers"
}